import typer
import uvicorn

from krypton_ml.core.loader.config import load_config
from krypton_ml.core.utils.logger import logger
from krypton_ml.core.models.cli_config import RootConfig
from krypton_ml.core.webserver.model_endpoints import load_model_endpoints
from krypton_ml.core.webserver.server_runtime import create_server_runtime

app = typer.Typer()

@app.command()
def main(config: str = typer.Argument(help="The config file")):
    parsed_config: RootConfig = load_config(config)
    app = create_server_runtime(parsed_config.krypton.server)
    app = load_model_endpoints(app, parsed_config.krypton.models)

    logger.info("Starting Krypton ML Server")
    uvicorn.run(app,
                host=parsed_config.krypton.server.host,
                port=parsed_config.krypton.server.port,
                log_config=None)

if __name__ == "__main__":
    app()


import json

from fastapi import FastAPI, Request
from fastapi.responses import JSONResponse

from langchain.load.dump import dumps
from krypton_ml.core.loader.module import load_module
from krypton_ml.core.models.cli_config import Model

model_registry = {}

def load_model_endpoints(app: FastAPI, models: [Model]):
    for idx, model in enumerate(models):
        lc_callable = load_module(model.module_path, model.callable)
        model_key = f"/{model.endpoint}"
        model_registry[model_key] = lc_callable

        @app.post(f"/{model.endpoint}", name=model.name, description=model.description, tags=model.tags)
        async def invoke_model(request: Request, input: dict):
            model_identifier = request.url.path
            response = model_registry[model_identifier].invoke(input)
            json_string = dumps(response, ensure_ascii=False)
            return JSONResponse(content={
                "response": json.loads(json_string)
            })

    return app



import importlib.util
import sys
import os
from typing import Callable

def load_module(module_path: str, callable_path: str) -> Callable:
    """
    Load a module from a file path and return a callable object.

    :param module_path: The file path to the Python module
    :param callable_path: The dot-separated path to the callable within the module
    :return: The callable object
    """

    # Split the callable path into package, module, and callable parts
    parts = callable_path.split('.')
    package_name = parts[0]
    module_name = parts[1]
    callable_name = '.'.join(parts[2:])

    # Construct the full path to the package
    package_path = os.path.join(module_path, package_name)

    # Add the package path to sys.path temporarily
    sys.path.insert(0, module_path)

    try:
        # Import the module
        module = importlib.import_module(f"{package_name}.{module_name}")

        # Get the callable from the module
        obj = module
        for part in parts[2:]:
            obj = getattr(obj, part)

        return obj

    except ImportError:
        raise ImportError(f"Could not import module {package_name}.{module_name} from {package_path}")
    except AttributeError:
        raise AttributeError(f"Could not find callable '{callable_name}' in module {package_name}.{module_name}")
    finally:
        # Remove the temporarily added path
        sys.path.pop(0)



import traceback

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse

from krypton_ml.core.models.cli_config import ServerConfig
from krypton_ml.core.utils.logger import logger
from krypton_ml.core.webserver.log_middleware import LogMiddleware


def create_server_runtime(server_config: ServerConfig):
    app = FastAPI(title="Krypton ML Server", version="0.1.0")

    # Add CORS middleware
    app.add_middleware(
        CORSMiddleware,
        allow_origins=server_config.allow_origins,
        allow_credentials=server_config.allow_credentials,
        allow_methods=server_config.allow_methods,
        allow_headers=server_config.allow_headers,
    )
    app.add_middleware(LogMiddleware)

    # Add a sample endpoint
    @app.get("/")
    async def root():
        return {"message": "Krpyton ML Server is running!"}

    # Health check endpoint
    @app.get("/health")
    async def health():
        return {"status": "ok"}

    # Error handling
    @app.exception_handler(Exception)
    async def generic_exception_handler(request, exc):
        exception_response = {
            "message": "An error occurred",
            "detail": str(exc),
        }
        if server_config.debug:
            exception_response["stack_trace"] = traceback.format_exc()

        return JSONResponse(
            status_code=500,
            content=exception_response,
        )

    return app


import json
import logging
from logging import Formatter

class JsonFormatter(Formatter):
    def __init__(self):
        super(JsonFormatter, self).__init__()

    def format(self, record):
        json_record = {}
        json_record["message"] = record.getMessage()
        if "req" in record.__dict__:
            json_record["req"] = record.__dict__["req"]
        if "res" in record.__dict__:
            json_record["res"] = record.__dict__["res"]
        if record.levelno == logging.ERROR and record.exc_info:
            json_record["err"] = self.formatException(record.exc_info)

        return json.dumps(json_record)

logger = logging.root
handler = logging.StreamHandler()
handler.setFormatter(JsonFormatter())
logger.handlers = [handler]
logger.setLevel(logging.DEBUG)

logging.getLogger("uvicorn.access").disabled = True


from pydantic_yaml import parse_yaml_raw_as

from krypton_ml.core.models.cli_config import RootConfig


def load_config(config: str) -> RootConfig:
    return parse_yaml_raw_as(RootConfig, open(config).read())


from typing import List, Optional

from pydantic import BaseModel


class ModelOptions(BaseModel):
    debug: bool = False

class Model(BaseModel):
    name: str
    type: str
    module_path: str
    callable: str
    endpoint: str
    options: Optional[ModelOptions] = None
    tags: List[str] = []
    description: str = ""

class ServerConfig(BaseModel):
    host: str = "0.0.0.0"
    port: int = 8000
    allow_origins: List[str] = []
    allow_credentials: bool = False
    allow_methods: List[str] = []
    allow_headers: List[str] = []
    debug: bool = False

class KryptonConfig(BaseModel):
    models: List[Model]
    server: Optional[ServerConfig]

class RootConfig(BaseModel):
    krypton: KryptonConfig


from starlette.middleware.base import BaseHTTPMiddleware

from krypton_ml.core.utils.logger import logger


class LogMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        response = await call_next(request)
        logger.info(
            "Incoming request",
            extra={
                "req": { "method": request.method, "url": str(request.url) },
                "res": { "status_code": response.status_code, },
            },
        )
        return response
